using System;
using System.ComponentModel;
using System.Collections;
using System.Diagnostics;

/**
 * TO DO : On text-change, complete box is parsed, otherwise pasting text is not colored properly
 *			scrollbars?
 */

namespace BrainfuckStudio
{
	/// <summary>
	/// Zusammenfassung fr BfSourceTextBox
	/// </summary>
	public class BfSourceTextBox : System.Windows.Forms.RichTextBox
	{
		const short WM_PAINT = 0x00F;
		const short WM_PASTE = 0x0302;

		private char[] KeyWords = {'<', '>', '.', ',', '+', '-', '[', ']'};

		public System.Windows.Forms.StatusBar statusBar;

		public static bool bReactOnChanging = true;
		public static bool bPaint = true;

		public BrainfuckStudio.Form1 parentForm = null;

		//public static bool bTextWasPasted = false;

		// the position in front of the next command to execute
		int SourcePos;

		private System.Drawing.Color HighlightColor = System.Drawing.Color.White;
		private System.Drawing.Color CommentColor = System.Drawing.Color.ForestGreen;

		//private System.Windows.Forms.Splitter splitter;
		//private System.Windows.Forms.Panel breakpointPanel;

		/// <summary>
		/// Constructor
		/// </summary>
		public BfSourceTextBox()
		{
			this.BackColor = System.Drawing.Color.Silver;
			
			// react on changing text
			this.TextChanged += new EventHandler(BfSourceTextBox_TextChanged);
			SourcePos = 0;

			this.SelectionChanged +=new EventHandler(BfSourceTextBox_SelectionChanged);

			// enable drag&drop handler
			this.DragEnter += new System.Windows.Forms.DragEventHandler(BfSourceTextBox_DragEnter);
			this.DragDrop += new System.Windows.Forms.DragEventHandler(BfSourceTextBox_DragDrop);

			// things for gui stuff
			SetStyle(System.Windows.Forms.ControlStyles.AllPaintingInWmPaint, true);
			SetStyle(System.Windows.Forms.ControlStyles.DoubleBuffer , true);

			// add contextmenu
			System.Windows.Forms.ContextMenu menu = new System.Windows.Forms.ContextMenu();
			menu.MenuItems.Add("Syntax coloring now!");
			menu.MenuItems[0].Click += new EventHandler(MenuSyntax_Click);
			this.ContextMenu = menu;
			
			
		}

		/// <summary>
		/// Returns true if keychar is a brainfuck command
		/// </summary>
		/// <param name="keychar"></param>
		/// <returns></returns>
		private bool isKeyWord(char keychar)
		{
			foreach (char c in KeyWords)
				if (c == keychar) return true;
			return false;
		}

		public void markNextCommand()
		{
			int pos = SourcePos;
			bool bIsKeyWord;
			while (pos < Text.Length)
			{
				/*if (Text[pos] == '\n')
				{
					//SelectionStart -= 1;
					SelectionLength = 0;
					this.ProcessCurrentLine(false);
					//SelectionStart += 1;
				}*/
				// ignore comments
				if (Text[pos] == ';')
				{
					while (pos < Text.Length && Text[pos] != '\n')
						pos++;
				}

				if (pos == Text.Length) break;

				bIsKeyWord = false;
				foreach (char c in KeyWords)
					if (c == Text[pos])
						bIsKeyWord = true;
				if (bIsKeyWord) break;
				pos++;
			}


			SelectionStart = pos;
			SelectionLength = 1;
			SelectionColor = System.Drawing.Color.LightPink;
		}

		public void unmarkPreviousCommand()
		{
			int pos = SourcePos;
			bool bIsKeyWord;
			while (pos >= 0)
			{
				bIsKeyWord = false;
				foreach (char c in KeyWords)
					if (c == Text[pos])
						bIsKeyWord = true;
				if (bIsKeyWord) break;
				pos--;
			}

			SelectionStart = pos-1;
			SelectionLength = 1;
			SelectionColor = System.Drawing.Color.White;
		}


		public void unmarkNextCommand()
		{
			int pos = SourcePos;
			bool bIsKeyWord;
			while (pos < Text.Length)
			{
				// ignore comments
				if (Text[pos] == ';')
				{
					while (pos < Text.Length && Text[pos] != '\n')
						pos++;
				}

				if (pos == Text.Length) break;

				bIsKeyWord = false;
				foreach (char c in KeyWords)
					if (c == Text[pos])
						bIsKeyWord = true;
				if (bIsKeyWord) break;
				pos++;
			}


			SelectionStart = pos;
			SelectionLength = 1;
			SelectionColor = System.Drawing.Color.White;
		}

		public void markCurrentCommand()
		{
			SelectionStart = SourcePos - 1;
			SelectionLength = 1;
			SelectionColor = System.Drawing.Color.LightPink;
		}

		public void unmarkCurrentCommand()
		{
			SelectionStart = SourcePos - 1;
			char s = Text[SourcePos - 1];
			SelectionLength = 1;
			SelectionColor = System.Drawing.Color.White;
		}


		/**
		 * Returns next char from source; returns 0 if there is no more command
		 **/
		public char getNextCommand()
		{
			while (SourcePos < Text.Length)
			{
				// ignore comments
				if (Text[SourcePos] == ';')
				{
					while (SourcePos < Text.Length && Text[SourcePos] != '\n')
						SourcePos++;
					if (SourcePos == Text.Length) break;
				}
				if (isKeyWord(Text[SourcePos++]))
					return Text[SourcePos-1];
			}
			return (char)0;
		}

		/**
		 * if SourcePos is at a '[', it moves SourcePos to the corressponding ']'
		 **/
		public bool goToClosingBracket()
		{
			char c;
			int brackets = 0;
			while (SourcePos < Text.Length)
			{
				c = Text[SourcePos++];
				if (c == '[')
					brackets++;
				if (c == ']')
				{
					if (brackets == 0) return true;
					if (brackets > 0)
					{
						brackets--;
						continue;
					}
					if (brackets < 0) return false;
				}
			}
			return false;
		}

		/**
		 * if SourcePos is at a ']', it moves SourcePos to the corressponding '['
		 **/
		public bool goToOpeningBracket()
		{
			char c;
			int brackets = 0;
			SourcePos--;
			while (SourcePos > 0)
			{
				c = Text[--SourcePos];
				if (c == ']')
					brackets++;
				if (c == '[')
				{
					if (brackets == 0) return true;
					if (brackets > 0)
					{
						brackets--;
						continue;
					}
					if (brackets < 0) return false;
				}
			}
			return false;
		}

		/**
		 * Returns true if source makes valid use of brackets, otherwise returns false
		 **/
		public bool CheckBrackets()
		{
			int brackets = 0;
			SourcePos = 0;
			while (SourcePos < Text.Length)
			{
				if (Text[SourcePos] == '[')
					brackets++;
				else if (Text[SourcePos] == ']')
					brackets--;
				SourcePos++;
			}
			return ((brackets == 0) ? true : false);
		}



		/**
		 * When text is changed, parse it for syntax highlighting
		 **/
		private void BfSourceTextBox_TextChanged(object sender, EventArgs e)
		{
			this.Modified = true;
			
			// don't do too much work
			if (!bReactOnChanging) return;
			if (this.TextLength == 0) return;
			bPaint = false;
			this.ProcessCurrentLine();	// THIS MUST BE STAY HERE!!!!
			bPaint = true;
		}

		/**
		 * called when cursor position in box changes
		 **/
		private void BfSourceTextBox_SelectionChanged(object sender, EventArgs e)
		{
			if (this.SelectionLength == 0)
				UpdateStatusBar();
		}

		/**
		 * Update statusbar of parent form with line and col number
		 **/
		private void UpdateStatusBar()
		{
			if (this.statusBar == null) return;

			int line, index, col;
			System.Drawing.Point pt;

			index = this.SelectionStart;
			line = this.GetLineFromCharIndex(index);
			pt = this.GetPositionFromCharIndex(index);
			pt.X = 0;
			col = index - this.GetCharIndexFromPosition(pt);

			this.statusBar.Panels[1].Text = "Line: " + line + "    Col: " + col;
		}


		/**
		 * Color complete box with brainfuck syntax;
		 * To do: speed improvements
		 **/
		public void ParseRichTextBox()
		{
			//this.ScrollBars = System.Windows.Forms.RichTextBoxScrollBars.None;
			int CurrentSelectionStart = this.SelectionStart;
			int CurrentSelectionLength = this.SelectionLength;

			// beginning make all black
			this.SelectionStart = 0;
			this.SelectionLength = Text.Length;
			this.SelectionColor = System.Drawing.Color.Black;

			int length = this.Text.Length;
			String s = this.Text;
			int i = 0;
			int start;
			while (i < length)
			{
				start = i;

				while ( (i < length) && (s[i] == '[' || s[i] == ']' || s[i] == '>' || s[i] == '<' ||
					s[i] == '+' || s[i] == '-' || s[i] == '.' || s[i] == ','))
					i++;

				if (i > start)
				{
					this.SelectionStart = start;
					this.SelectionLength = i - start;
					this.SelectionColor = HighlightColor;
				}
				else
					i++;
			}

			// handle comments
			i = 0;
			while (i < length)
			{
				if (s[i] == ';')
				{
					start = i;
					// find end of line
					while (( i < length) && (s[i] != '\n'))
						i++;

					if (i > start)
					{
						this.SelectionStart = start;
						this.SelectionLength = i - start;
						this.SelectionColor = CommentColor;
					}
				}
				i++;
			}

			if (CurrentSelectionStart >=0)
			{
				this.Select(CurrentSelectionStart, CurrentSelectionLength);
				this.SelectionColor = System.Drawing.Color.Black;
			}
		}


		
		/**
		 *  Color syntax of line where cursor resides
		 **/
		public void ProcessCurrentLine()
		{
			int CurrentSelectionStart = this.SelectionStart;
			int CurrentSelectionLength = this.SelectionLength;

			
			

			// find start of line
			int pos = CurrentSelectionStart;
			while ( (pos > 0) && (this.Text[pos-1] != '\n'))
				pos--;

			int pos2 = CurrentSelectionStart;
			while ( (pos2 < this.Text.Length) && (this.Text[pos2] != '\n') )
				pos2++;

			// TEST :: make line black
			//this.SelectionColor = System.Drawing.Color.Black;


			// length of line
			int numOfChars = pos2 - pos;
			String s = this.Text.Substring(pos, pos2 - pos);

			int i = 0;
			int start;
			try
			{
				while (i < numOfChars)
				{
					start = i;

					while ( (i < numOfChars) && (s[i] == '[' || s[i] == ']' || s[i] == '>' || s[i] == '<' ||
						s[i] == '+' || s[i] == '-' || s[i] == '.' || s[i] == ','))
						i++;

					if (i > start)		// we found keywords, so color them
					{
						this.SelectionStart = pos + start;
						this.SelectionLength = i - start;
						this.SelectionColor = HighlightColor;
					}
					else
						i++;
				}
			}
			catch (System.IndexOutOfRangeException) { }

			// handle comments
			i = 0;
			start = 0;
			while (i < numOfChars)
			{
				if (s[i] == ';')
				{
					start = i;
					// find end of line
					while (( i < numOfChars) && (s[i] != '\n'))
						i++;

					if (i > start)
					{
						this.SelectionStart = pos + start;
						this.SelectionLength = i - start;
						this.SelectionColor = CommentColor;
					}
				}
				i++;
			}


			this.SelectionLength = 0;

			if (CurrentSelectionStart >=0)
			{
				this.Select(CurrentSelectionStart, CurrentSelectionLength);
				this.SelectionColor = System.Drawing.Color.Black;
			}

		}
		


		/**
		 * Load a source file into the box and parse it
		 **/
		public void LoadSourceFile(String filename, System.Windows.Forms.RichTextBoxStreamType rtst)
		{
			bPaint = false;
			bReactOnChanging = false;
			this.LoadFile(filename, rtst);
			ParseRichTextBox();				
			bReactOnChanging = true;
			bPaint = true;
		}

		/**
		 * Save source to file
		 **/
		public bool SaveSourceFile(String filename, System.Windows.Forms.RichTextBoxStreamType rtst)
		{
			this.SaveFile(filename, rtst);
			return true;
		}

		/**
		 * Reset box for execution
		 **/
		public void PrepeareForExecution()
		{
			SourcePos = 0;
			SelectionStart = 0;
			SelectionLength = 0;
		}

		/**
		 * Context Menu handling; just parse complete box
		 **/
		private void MenuSyntax_Click(object sender, EventArgs e)
		{
			bPaint = false;
			bReactOnChanging = false;
			this.ParseRichTextBox();
			bReactOnChanging = true;
			bPaint = true;
			this.Invalidate();
		}

		#region AvoidFlicker
		/// <summary>
		/// Handle WM_PAINT message ourselves to reduce flickering
		/// Taken from: http://www.c-sharpcorner.com/UploadFile/mgold/SimpleColorSyntaxCodeEditorPartI11212005052757AM/SimpleColorSyntaxCodeEditorPartI.aspx?ArticleID=2e3d0967-802e-4d2c-ac5a-6da11a47f90f
		/// </summary>
		/// <param name="m"></param>
		protected override void WndProc(ref System.Windows.Forms.Message m)
		{
			// Code courtesy of Mark Mihevc
			// sometimes we want to eat the paint message so we don't have to see all the
			// flicker from when we select the text to change the color.
			if (m.Msg == WM_PAINT)
			{
				if (bPaint)
					base.WndProc(ref m);
				else
					m.Result = IntPtr.Zero;
			}
			//else if (m.Msg == WM_PASTE)
			//	bTextWasPasted = true;
			else
				base.WndProc (ref m);
		}

		#endregion

		#region Properties

		public int Position
		{
			get { return this.SourcePos; }
		}

		#endregion

		#region Drag and Drop
		/**
		 * Handle drag & drop. Make sure it's an file.
		 **/
		private void BfSourceTextBox_DragEnter(object sender, System.Windows.Forms.DragEventArgs e)
		{
			if (e.Data.GetDataPresent(System.Windows.Forms.DataFormats.FileDrop, false))
			{
				e.Effect = System.Windows.Forms.DragDropEffects.All;
			}
		}

		/**
		 * Process dragged file.
		 **/
		private void BfSourceTextBox_DragDrop(object sender, System.Windows.Forms.DragEventArgs e)
		{
			// dragged item is a file, so get it's filename.
			// note we just process the first one! If more than one file is dropped, the other are ignored.
			string[] filenames = (string[])e.Data.GetData(System.Windows.Forms.DataFormats.FileDrop);

			if (this.parentForm == null) return;
			parentForm.CloseFile();

			// if file was really closed, open the dragged one
			if (!parentForm.isFileOpened)
			{
				parentForm.LoadFile(filenames[0]);
			}
			
		}

		#endregion


		protected override void OnResize(EventArgs e)
		{
			
		}

	}
}
