using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using Microsoft.Win32;
using System.Runtime.InteropServices;

namespace BrainfuckStudio
{
	/// <summary>
	/// Zusammenfassung fr FileExtensionDlg.
	/// </summary>
	public class FileExtensionDlg : System.Windows.Forms.Form
	{
		private System.Windows.Forms.Button OKButton;
		/// <summary>
		/// Erforderliche Designervariable.
		/// </summary>
		private System.ComponentModel.Container components = null;
		private System.Windows.Forms.Button CancButton;
		private System.Windows.Forms.CheckBox bCheckBox;
		private System.Windows.Forms.CheckBox bfCheckBox;
		private System.Windows.Forms.CheckBox bfsCheckBox;

		// *** OWN VARIABLES ***
		private static string Description = "Brainfuck Source File";

		// we need SHChangeNotify to tell explorer that we changed a file association
		[DllImport("shell32.dll")]
		public extern static void SHChangeNotify(long eventId, uint flags, IntPtr item1, IntPtr item2);
		private const long SHCNE_ASSOCCHANGED = 0x08000000;
		private const uint SHCNF_IDLIST = 0;

		// save which extensions are associated
		bool bExt, bfExt ,bfsExt;

		/**
		 * constructor
		 **/
		public FileExtensionDlg()
		{
			InitializeComponent();
		}


		/// <summary>
		/// Die verwendeten Ressourcen bereinigen.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Vom Windows Form-Designer generierter Code
		/// <summary>
		/// Erforderliche Methode fr die Designeruntersttzung. 
		/// Der Inhalt der Methode darf nicht mit dem Code-Editor gendert werden.
		/// </summary>
		private void InitializeComponent()
		{
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(FileExtensionDlg));
			this.bCheckBox = new System.Windows.Forms.CheckBox();
			this.bfCheckBox = new System.Windows.Forms.CheckBox();
			this.bfsCheckBox = new System.Windows.Forms.CheckBox();
			this.OKButton = new System.Windows.Forms.Button();
			this.CancButton = new System.Windows.Forms.Button();
			this.SuspendLayout();
			// 
			// bCheckBox
			// 
			this.bCheckBox.Location = new System.Drawing.Point(8, 8);
			this.bCheckBox.Name = "bCheckBox";
			this.bCheckBox.Size = new System.Drawing.Size(144, 24);
			this.bCheckBox.TabIndex = 0;
			this.bCheckBox.Text = "associate with *.b files";
			// 
			// bfCheckBox
			// 
			this.bfCheckBox.Location = new System.Drawing.Point(8, 32);
			this.bfCheckBox.Name = "bfCheckBox";
			this.bfCheckBox.Size = new System.Drawing.Size(144, 24);
			this.bfCheckBox.TabIndex = 1;
			this.bfCheckBox.Text = "associate with *.bf files";
			// 
			// bfsCheckBox
			// 
			this.bfsCheckBox.Location = new System.Drawing.Point(8, 56);
			this.bfsCheckBox.Name = "bfsCheckBox";
			this.bfsCheckBox.Size = new System.Drawing.Size(144, 24);
			this.bfsCheckBox.TabIndex = 2;
			this.bfsCheckBox.Text = "associate with *.bfs files";
			// 
			// OKButton
			// 
			this.OKButton.Location = new System.Drawing.Point(24, 104);
			this.OKButton.Name = "OKButton";
			this.OKButton.TabIndex = 3;
			this.OKButton.Text = "OK";
			this.OKButton.Click += new System.EventHandler(this.OKButton_Click);
			// 
			// CancButton
			// 
			this.CancButton.Location = new System.Drawing.Point(136, 104);
			this.CancButton.Name = "CancButton";
			this.CancButton.TabIndex = 4;
			this.CancButton.Text = "Cancel";
			this.CancButton.Click += new System.EventHandler(this.CancButton_Click);
			// 
			// FileExtensionDlg
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(248, 136);
			this.Controls.Add(this.CancButton);
			this.Controls.Add(this.OKButton);
			this.Controls.Add(this.bfsCheckBox);
			this.Controls.Add(this.bfCheckBox);
			this.Controls.Add(this.bCheckBox);
			this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
			this.MaximizeBox = false;
			this.MinimizeBox = false;
			this.Name = "FileExtensionDlg";
			this.Text = "File Extension Associations";
			this.Load += new System.EventHandler(this.FileExtensionDlg_Load);
			this.ResumeLayout(false);

		}
		#endregion


		

		/*********************************************************************
		 * REGISTRY HELPER FUNCTIONS
		 ********************************************************************/

		/**
		 * Register specified file extension with this application
		 * description is the explaining string in explorer 
		 **/
		public static void RegisterFileExtension(string extension)
		{
			// make sure we have the dot
			if (extension[0] != '.')
				extension = '.' + extension;

			RegistryKey Key = Registry.ClassesRoot.OpenSubKey(extension, true);
			if (Key != null)
			{
				// extension is already associated!
				// check if it's this programm or another
				String desc = Key.GetValue("").ToString();
				if (desc != Description)		// associated to another programm
				{
					System.Windows.Forms.DialogResult Dlgresult;
					Dlgresult = System.Windows.Forms.MessageBox.Show("The file extension " + extension + " is already associated to" + 
						"a programm. \n Do you really want to associate it with Brainfuck Sutdio?",
						"Warning!", 
						System.Windows.Forms.MessageBoxButtons.YesNo,
						System.Windows.Forms.MessageBoxIcon.Warning);
					if (Dlgresult == System.Windows.Forms.DialogResult.No)
					{
						Key.Close();
						return;
					}
				}
			}
			else 
			{
				Registry.ClassesRoot.CreateSubKey(extension);
				Key = Registry.ClassesRoot.OpenSubKey(extension, true);
			}

			Key.SetValue("", Description);
			Key.Close();
		}

		/**
		 * The last necessary registry is created and explorer is notified
		 **/
		public static void FinishRegisterAndNotify()
		{
			// ok, do the real key
			RegistryKey Key = Registry.ClassesRoot.CreateSubKey(Description);
			Key.SetValue("", Description);
			// assign the icon
			RegistryKey SubKey = Key.CreateSubKey("defaulticon");
			SubKey.SetValue("", System.Windows.Forms.Application.ExecutablePath);
			// link the with exetutable	
			SubKey = Key.CreateSubKey("shell");
			SubKey = SubKey.CreateSubKey("open");
			SubKey = SubKey.CreateSubKey("command");			
			SubKey.SetValue("", "\"" + System.Windows.Forms.Application.ExecutablePath + "\" \"" +  "%1" + "\"");
			SubKey.Close();
			Key.Close();

			// tell the system that we changed a file association
			SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, IntPtr.Zero, IntPtr.Zero);
		}


		/**
		 * Unregister the file extension in the system.
		 * TO DO: possible associations of other applications are also deleted.
		 **/
		public static void UnregisterFileExtension(string extension)
		{
			// make sure we have the dot
			if (extension[0] != '.')
				extension = '.' + extension;
			RegistryKey Key = Registry.ClassesRoot.OpenSubKey(extension, true);
			if (Key != null)
			{
				// extension is already associated!
				// check if it's this programm or another
				String desc = Key.GetValue("").ToString();
				if (desc != Description)		// associated to another programm
				{
					System.Windows.Forms.DialogResult Dlgresult;
					Dlgresult = System.Windows.Forms.MessageBox.Show("The file extension " + extension + " is associated to" + 
						"another  programm. \n Do you really want to delete it?",
						"Warning!", 
						System.Windows.Forms.MessageBoxButtons.YesNo,
						System.Windows.Forms.MessageBoxIcon.Warning);
					if (Dlgresult == System.Windows.Forms.DialogResult.No)
					{
						Key.Close();
						return;
					}
				}
			}
			else return;

			Registry.ClassesRoot.DeleteSubKeyTree(extension);
			Key.Close();
			SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, IntPtr.Zero, IntPtr.Zero);
		}

		/**
		 * Finish unregistering by deleting the last necessary reg key
		 **/
		public static void FinishUnregister()
		{
			RegistryKey Key = Registry.ClassesRoot.OpenSubKey(Description);
			if (Key == null) return;
			Registry.ClassesRoot.DeleteSubKeyTree(Description);
			Key.Close();
			// tell the system that we changed a file association
			SHChangeNotify(SHCNE_ASSOCCHANGED, SHCNF_IDLIST, IntPtr.Zero, IntPtr.Zero);
		}

		/**
		 * Check if given extension is associated to 'Brainfuck Source File'
		 **/
		public static bool IsRegisteredToUs(string extension)
		{
			// make sure we have the dot
			if (extension[0] != '.')
				extension = '.' + extension;
			RegistryKey Key = Registry.ClassesRoot.OpenSubKey(extension, true);
			if (Key == null) return false;
			String desc = Key.GetValue("").ToString();
			if (desc != Description) return false;
			Key.Close();
			return true;
		}

		private void FileExtensionDlg_Load(object sender, System.EventArgs e)
		{
			if (IsRegisteredToUs(".b")) 
				bCheckBox.Checked = true; else bCheckBox.Checked = false;
			if (IsRegisteredToUs(".bf")) 
				bfCheckBox.Checked = true; else bfCheckBox.Checked = false;
			if (IsRegisteredToUs(".bfs")) 
				bfsCheckBox.Checked = true; else bfsCheckBox.Checked = false;
			bExt = bCheckBox.Checked;
			bfExt = bfCheckBox.Checked;
			bfsExt = bfsCheckBox.Checked;
		}

		private void CancButton_Click(object sender, System.EventArgs e)
		{
			Close();
		}

		private void OKButton_Click(object sender, System.EventArgs e)
		{
			bool Changed = false;
			if (bExt != bCheckBox.Checked)
			{
				Changed = true;
				if (bCheckBox.Checked)
					RegisterFileExtension(".b");
				else UnregisterFileExtension(".b");
			}
			if (bfExt != bfCheckBox.Checked)
			{
				Changed = true;
				if (bfCheckBox.Checked)
					RegisterFileExtension(".bf");
				else UnregisterFileExtension(".bf");
			}
			if (bfsExt != bfsCheckBox.Checked)
			{
				Changed = true;
				if (bfsCheckBox.Checked)
					RegisterFileExtension(".bfs");
				else UnregisterFileExtension(".bfs");
			}

			if (Changed)
			{
				if (!bCheckBox.Checked && !bfCheckBox.Checked && !bfsCheckBox.Checked)
					FinishUnregister();
				else
					FinishRegisterAndNotify();
			}

			Close();
		}
	}
}
