using System;
using System.ComponentModel;
using System.Collections;
using System.Diagnostics;
using System.Drawing;

namespace BrainfuckStudio
{
	/// <summary>
	/// Zusammenfassung fr VisualMemPanel.
	/// </summary>
	public class VisualMemPanel : System.Windows.Forms.Panel, IObserver
	{
		// helper variables to avoid flicker
		const short WM_PAINT = 0x00F;
		public static bool bPaint = true;

		public const int CELLWIDTH = 50;
		public const int CELLWIDTH2 = CELLWIDTH / 2;
		public const int CELLHEIGHT = 20;
		public const int CELLHEIGHT2 = CELLHEIGHT / 2;
		private const int SCROLLBARDISTANCE2 = 50;
		
		// the y-value of the memoryband 
		public int CellsY;

		// reference to the memory
		private MemoryBand	Memory;


		// ** some variables for drawing

		// number of cells we can draw/see, depending on width of our panel
		private	int		numOfCells;
		// first cell we currently see
		private int		firstVisibleCell;
		// last cell we currently see
		private int		lastVisibleCell;

		// if signal for repaint came from the memory band, it's true, otherwise false.
		// Helps for deciding if we should scroll the panel by code
		private bool	bRepaintFromMem = false;

		System.Windows.Forms.HScrollBar scrollbar = new System.Windows.Forms.HScrollBar();

		/**
		 * normal constructor
		 **/
		public VisualMemPanel()
		{
			// start at first cell
			this.firstVisibleCell = 0;
			// give it some dummy value, it's set on resize
			this.lastVisibleCell = 10;

			scrollbar.Parent = this;
			scrollbar.Location = new Point(0, this.Height - System.Windows.Forms.SystemInformation.HorizontalScrollBarHeight
											- System.Windows.Forms.SystemInformation.Border3DSize.Height - 5);
			scrollbar.Size = new Size(this.Width, System.Windows.Forms.SystemInformation.HorizontalScrollBarHeight);
			scrollbar.Minimum = 0;
			scrollbar.SmallChange = 1;
			scrollbar.LargeChange =  5;
			scrollbar.Scroll += new System.Windows.Forms.ScrollEventHandler(OnScrolling);
			scrollbar.BringToFront();
		}

		/**
		 * Repaint on scrolling
		 **/
		void OnScrolling(Object obj, System.Windows.Forms.ScrollEventArgs sea)
		{
			this.Invalidate();
		}

		
		/**
		 * Paint the panel; here is the whole action
		 **/
		protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
		{
			// call base paint
			base.OnPaint (e);

			// if we have no reference to memory, bye
			if (Memory == null) return;

			// clear background
			e.Graphics.Clear(this.BackColor);

			int curPos = Memory.CurrentPos;
			int firstCell, lastCell;

			// the repaint signal came from the memory band, so take care if we have to scroll ourselves
			if (bRepaintFromMem)
			{
				// check if we are not outside the visible area
				if (curPos < this.firstVisibleCell || curPos > this.lastVisibleCell)
				{
					// current pos not visible any more, so calc new visible part from mempanel
					// and update scrollbar
					int temp = curPos - this.numOfCells / 2;
					if (temp < 0) temp = 0;
					if (temp > Memory.MemLength) temp = Memory.MemLength - 1;
					scrollbar.Minimum = Math.Max(0, temp - VisualMemPanel.SCROLLBARDISTANCE2);
					scrollbar.Maximum = Math.Min(Memory.MemLength, temp + 
						VisualMemPanel.SCROLLBARDISTANCE2);
					scrollbar.Value = curPos;
					this.firstVisibleCell = temp;
					this.lastVisibleCell = temp + this.numOfCells;
				}
				bRepaintFromMem = false;
				firstCell = this.firstVisibleCell; 
				lastCell = this.lastVisibleCell; 
			}
			// repaint signal is not from the memory band, so support manual scrolling
			else
			{
				firstCell = scrollbar.Value;
				lastCell = firstCell + numOfCells;
			}

			// white background
			Brush whiteBrush = Brushes.White;
			int upperbound = CellsY - CELLHEIGHT2;
			int lowerbound = CellsY + CELLHEIGHT2;

			e.Graphics.FillRectangle(whiteBrush, 0, upperbound,
				this.Width, CELLHEIGHT);
			
			// black thin pen
			System.Drawing.Pen pen = new System.Drawing.Pen(Color.Black, 1.0f);
			// paint all black lines
			e.Graphics.DrawLine(pen, 0, upperbound, this.Width, upperbound);
			e.Graphics.DrawLine(pen, 0, lowerbound, this.Width, lowerbound);
			for (int i = 0; i <= numOfCells; i++)
				e.Graphics.DrawLine(pen, i * CELLWIDTH, upperbound,
					i * CELLWIDTH, lowerbound);

			// highlight current position in yellow
			e.Graphics.FillRectangle(Brushes.Yellow, 
				(Memory.CurrentPos - firstCell) * CELLWIDTH+1, upperbound + 1,
				CELLWIDTH-1, CELLHEIGHT-1);
		
			FillVisibleValues(firstCell, lastCell, e);
			e.Graphics.Flush();						
		}	
		
		
		/**
		 * Write the number of the cells and the values in the cells
		 */
		protected void FillVisibleValues(int startCell, int endCell, System.Windows.Forms.PaintEventArgs e)
		{
			System.Drawing.Font font = new System.Drawing.Font("Arial", 10);
			System.Drawing.Font font2 = new System.Drawing.Font("Arial", 7);
			SizeF textSize;
			String s;
			for (int i = startCell; i <= endCell; i++)
			{
				s = "";
				// BAD STYLE - prevent array exception - to improve!
				if (i >= Memory.MemLength) break;
				// value
				int val = Memory.Memory[i];
				textSize = e.Graphics.MeasureString(val.ToString(), font);
				e.Graphics.DrawString(val.ToString(), font, System.Drawing.Brushes.Black, 
					CELLWIDTH * (i - startCell) + CELLWIDTH2 - textSize.Width / 2, 
					CellsY - textSize.Height / 2);	
				// cell number
				textSize = e.Graphics.MeasureString(i.ToString(), font2);
				e.Graphics.DrawString(i.ToString(), font2, System.Drawing.Brushes.Black, 
					CELLWIDTH * (i - startCell) + CELLWIDTH2 - textSize.Width / 2, 
					CellsY - CELLHEIGHT2 - textSize.Height);
				// hex value
				if (Memory.DataType == MemoryBand.DATATYPE.SIGNED_SHORT)
					s = String.Format("{0:X4}",(short)val);
				else if (Memory.DataType == MemoryBand.DATATYPE.SIGNED_BYTE)
					s = String.Format("{0:X2}",(sbyte)val);
				else
					s = String.Format("{0:X}",val);
				textSize = e.Graphics.MeasureString(s, font2);
				e.Graphics.DrawString(s, font2, System.Drawing.Brushes.Black, 
					CELLWIDTH * (i - startCell) + CELLWIDTH2 - textSize.Width / 2, 
					CellsY + CELLHEIGHT2 + textSize.Height / 2);
				// char
				e.Graphics.DrawString(((char)val).ToString(), font2, System.Drawing.Brushes.Black, 
					CELLWIDTH * (i - startCell) + CELLWIDTH2 - textSize.Width / 2, 
					CellsY + CELLHEIGHT + CELLHEIGHT2 + textSize.Height / 2);

			}
		}
		
		/**
		 * Panel was resized
		 **/
		protected override void OnResize(EventArgs e)
		{
			base.OnResize (e);
			
			numOfCells = this.Width / CELLWIDTH;	
			this.lastVisibleCell = this.firstVisibleCell  + this.numOfCells;
		
			scrollbar.Size = new Size(this.Width - System.Windows.Forms.SystemInformation.BorderSize.Width
				- System.Windows.Forms.SystemInformation.Border3DSize.Width , 
				System.Windows.Forms.SystemInformation.HorizontalScrollBarHeight);
			scrollbar.Location = new Point(0, this.Height - System.Windows.Forms.SystemInformation.HorizontalScrollBarHeight
				- System.Windows.Forms.SystemInformation.Border3DSize.Height-1);
			CellsY = 40;
		}	


		// ** Properties **
		#region properties

		public MemoryBand Memoryband
		{
			set 
			{
				Memory = value;
				this.scrollbar.Maximum = 100;
			}
		}
		#endregion

		#region IObserver Member

		public void Update(ISubject sender)
		{
			bRepaintFromMem = true;		// scroll automatic if necessary
			this.Invalidate();
			
		}

		#endregion

		#region AvoidFlicker
		/// <summary>
		/// Handle WM_PAINT message ourselves to reduce flickering
		/// Taken from: http://www.c-sharpcorner.com/UploadFile/mgold/SimpleColorSyntaxCodeEditorPartI11212005052757AM/SimpleColorSyntaxCodeEditorPartI.aspx?ArticleID=2e3d0967-802e-4d2c-ac5a-6da11a47f90f
		/// </summary>
		/// <param name="m"></param>
		protected override void WndProc(ref System.Windows.Forms.Message m)
		{
			// Code courtesy of Mark Mihevc
			// sometimes we want to eat the paint message so we don't have to see all the
			// flicker from when we select the text to change the color.
			if (m.Msg == WM_PAINT)
			{
				if (bPaint)
					base.WndProc(ref m); 
				else
					m.Result = IntPtr.Zero; 
			}
			else
				base.WndProc (ref m);
		}

		#endregion
	}
}
